/*=============================================================================
  Copyright (c) 2001-2008 Joel de Guzman
  Copyright (c) 2001-2008 Hartmut Kaiser
  http://spirit.sourceforge.net/

  Distributed under the Boost Software License, Version 1.0. (See accompanying
  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
=============================================================================*/
#ifndef BOOST_SPIRIT_INCLUDE_CLASSIC_FUNCTOR_PARSER
#define BOOST_SPIRIT_INCLUDE_CLASSIC_FUNCTOR_PARSER
#include <boost/spirit/home/classic/utility/functor_parser.hpp>
#endif
